@import url("resource:///plugins/color-picker/themes/shared.css");

/*
 * FileChooserDialog preview
 */

gstylecolorpanel gstylepalettewidget.dnd {
  border-color: #4e9a06;
  background-color: #4e9a06;
}
