@import url("resource:///org/gnome/builder/themes/shared.css");

/* Darker grey accents used throughtout */
@define-color theme_accent_color #858c98;
@define-color theme_accent_bg_color #353945;
/*@define-color theme_accent_unfocused_color #89929e;
@define-color theme_accent_bg_unfocused_color #313843;*/
@define-color theme_button_hover_bg_color #454C5C;
@define-color theme_button_hover_border_color #262932;

surfaceswitcher {
  background-color: @theme_accent_bg_color;
  border-top: 1px solid @borders;
  border-right: 1px solid @borders;
}

surfaceswitcher button {
  color: @theme_accent_color;
  background-color: @theme_accent_bg_color;
  border-radius: 3px;
  box-shadow: none;
  border: none;
  margin: 1px;
}

surfaceswitcher button:hover {
  border-color: @theme_button_hover_border_color;
  background-color: @theme_button_hover_bg_color;
}

surfaceswitcher button:checked {
  color: white;
  background-color: @wm_button_active_bg;
}

surfaceswitcher button:checked:backdrop {
  color: #c2c4c7;
}


panel stackswitcher button {
  color: @theme_fg_color;
  background-color: transparent;
  border: none;
}
panel stackswitcher button:checked {
  color: @theme_selected_bg_color;
}
/* All boxes */
panel > box > box.horizontal > stackswitcher > button:hover {
  border: 1px solid @borders;
}
/* Box above file switcher */
panel > box.vertical:first-child > box.horizontal {
  border: 1px solid @borders;
}


/* Builder pane */
window.workspace buildsurface list {
  border-right: 1px solid @borders;
  background-color: @theme_base_color;
}
window.workspace buildsurface list row {
  padding: 10px;
  border-bottom: 1px solid alpha(@borders, 0.50);
}
window.workspace buildsurface list row:last-child {
  border-bottom: none;
}


/* omnibar popover, remove popover padding */
popover.omnibar > * > * {
  margin: 0;
  padding: 0;
}


/* utilities stack switcher */
ideeditorutilities > dzldockpaned > box > stackswitcher {
  margin: 6px;
}
