popover.symbols-button {
  padding: 18px 6px 6px 6px;
}
popover.symbols-button entry {
  margin: 0 8px;
}
popover.symbols-button treeview {
  -GtkTreeView-expander-size: 0;
  -GtkTreeView-horizontal-separator: 0;
  -GtkTreeView-vertical-separator: 6;
  -gtk-icon-source: none;
  background: transparent;
  color: @theme_fg_color;
}
popover.symbols-button treeview:hover {
  background-color: alpha(@theme_fg_color, 0.05);
}
popover.symbols-button treeview:backdrop {
  color: @theme_unfocused_fg_color;
}
