/* Original Stylesheet */
@import url("https://media.readthedocs.org/css/sphinx_rtd_theme.css");

/* Source Sans Pro and Source Code Pro */
@import url("https://fonts.googleapis.com/css?family=Source+Sans+Pro:700,600,700i,400,400i,300,200&amp;subset=latin-ext");
@import url('https://fonts.googleapis.com/css?family=Source+Code+Pro');

/* Various GNOME Styles and Colors */

body {
  font-family: "Source Sans Pro", sans-serif;
  font-size: 12pt;
  font-weight: 400;
  color: #333333;
}

h1 {
  font-family: "Source Sans Pro", sans-serif;
  font-size: 38pt;
  font-weight: 200;
  color: #333333;
}

h2 {
  font-family: "Source Sans Pro", sans-serif;
  font-size: 24pt;
  font-weight: 400;
  color: #4a86cf;
}

h3 {
  font-family: "Source Sans Pro", sans-serif;
  font-size: 18pt;
  font-weight: 300;
  color: #333333;
}

h4 {
  font-family: "Source Sans Pro", sans-serif;
  font-size: 12pt;
  font-weight: 600;
  color: #808080;
}

p {
  font-family: "Source Sans Pro", sans-serif;
  font-size: 12pt;
  font-weight: 400;
  color: #333333;
}

.linenos .linenodiv pre,
.highlight pre {
  font-family: "Source Code Pro", monospace;
  font-size: 10pt !important;
  line-height: 1.2em !important;
}

table.column1-code tr td:first-child,
table.column2-code tr td:nth-child(2) {
  font-family: "Source Code Pro", monospace;
  font-size: 9pt;
}
table.column1-right tr td:first-child {
  text-align: right;
}

span.versionmodified {
  color: #01808e;
}

span.deprecated {
  color: #f57900;
}

span.std.std-ref {
  font-family: "Source Code Pro", monospace;
  font-size: 10pt;
}

div.admonition.note {
  background-color: #4a86cf33; }
div.admonition.note .first {
  background-color: #4a86cf; }

div.admonition.tip {
  background-color: #01808e33; }
div.admonition.tip .first {
  background-color: #01808e; }

div.admonition.warning {
  background-color: #f5790033; }
div.admonition.warning .first {
  background-color: #f57900; }

a.icon-home, a.icon-home:hover {
  display: inline-block;
  padding: 2px 2px 2px 20px;
  background: transparent url(img/org.gnome.Builder.png) center left no-repeat;
}
.fa-home::before, .icon-home::before {
  content: "";
}

.wy-side-nav-search {
  background-color: #4a86cf;
}

a, a:visited { color: #4a86cf; }
  .wy-menu a { color: #808080; }
  .wy-side-nav-search a { color: #fff; }
